import json
import os
from typing import Optional
from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import cluster

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.command_env.config_http_corosync import (
    corosync_running_check_response,
)

from .common import corosync_conf_fixture, node_fixture

_FIXTURE_NEW_NAME = "new"
_FIXTURE_GFS2_INVALID_NAME = "a.b"

_FIXTURE_GENERIC_FS_RESOURCE = """
    <primitive id="FS" class="ocf" provider="heartbeat" type="Filesystem">
        <instance_attributes id="fs-instance_attributes">
            <nvpair id="fs-fstype" name="fstype" value="xfs"/>
            <nvpair id="bogus" name="attribute" value="gfs2"/>
        </instance_attributes>
    </primitive>
"""

_FIXTURE_GFS2_FS_RESOURCE = """
    <primitive id="GFS" class="ocf" provider="heartbeat" type="Filesystem">
        <instance_attributes id="fs-instance_attributes">
            <nvpair id="gfs-fstype" name="fstype" value="gfs2"/>
        </instance_attributes>
    </primitive>
"""

_FIXTURE_DLM_RESOURCE = """
    <primitive id="D" class="ocf" provider="pacemaker" type="controld"/>
"""


class FixtureMixin:
    node_labels = ["node-1", "node-2"]

    @staticmethod
    def fixture_corosync_conf(cluster_name: Optional[str] = None) -> str:
        return corosync_conf_fixture(
            node_list=[node_fixture("node-1", 1), node_fixture("node-2", 2)],
            cluster_name=cluster_name,
        )

    def fixture_remove_name_prop_call(
        self, communication_list=None, output=None
    ):
        self.config.http.place_multinode_call(
            "cluster.remove-name",
            communication_list=(
                communication_list
                if communication_list
                else [{"label": node} for node in self.node_labels]
            ),
            action="api/v1/cluster-property-remove-name/v1",
            raw_data=json.dumps({}),
            output=(
                output
                if output
                else json.dumps(
                    {
                        "status": "success",
                        "status_msg": None,
                        "report_list": [],
                        "data": "",
                    }
                )
            ),
        )

    def fixture_corosync_offline_check_reports(self):
        reports = [
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED)
        ]
        reports.extend(
            fixture.info(
                report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_STOPPED, node=n
            )
            for n in self.node_labels
        )
        return reports

    def fixture_remove_name_prop_reports(self):
        reports = [fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED)]
        reports.extend(
            fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVED, node=n)
            for n in self.node_labels
        )
        return reports

    def _offline_node_communication_list(self):
        return [
            {
                "label": self.node_labels[0],
                "response_code": 400,
                "output": "Fail",
            }
        ] + [{"label": node} for node in self.node_labels[1:]]


class RenameCluster(FixtureMixin, TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.cib_path = os.path.join(settings.cib_dir, "cib.xml")
        self.runner_env = {"CIB_file": self.cib_path}
        self.config.env.set_known_nodes(self.node_labels)

    def test_success(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call()
        self.config.env.push_corosync_conf(
            corosync_conf_text=self.fixture_corosync_conf(_FIXTURE_NEW_NAME),
            need_stopped_cluster=True,
        )

        cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_no_cib(self):
        self.config.fs.exists(self.cib_path, False)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call()
        self.config.env.push_corosync_conf(
            corosync_conf_text=self.fixture_corosync_conf(_FIXTURE_NEW_NAME),
            need_stopped_cluster=True,
        )

        cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_invalid_name(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), "")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="cluster name",
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ]
        )

    def test_gfs2_invalid_name(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(
                self.env_assist.get_env(), _FIXTURE_GFS2_INVALID_NAME
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2,
                    force_code=report_codes.FORCE,
                    cluster_name=_FIXTURE_GFS2_INVALID_NAME,
                    max_length=32,
                    allowed_characters="a-z A-Z 0-9 _-",
                )
            ]
        )

    def test_gfs2_invalid_name_forced(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call()
        self.config.env.push_corosync_conf(
            corosync_conf_text=self.fixture_corosync_conf(
                _FIXTURE_GFS2_INVALID_NAME
            ),
            need_stopped_cluster=True,
        )

        cluster.rename(
            self.env_assist.get_env(),
            _FIXTURE_GFS2_INVALID_NAME,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2,
                    cluster_name=_FIXTURE_GFS2_INVALID_NAME,
                    max_length=32,
                    allowed_characters="a-z A-Z 0-9 _-",
                )
            ]
            + self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_node_offline_corosync_check(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(
            communication_list=self._offline_node_communication_list()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            [
                fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_labels[0],
                    command="remote/status",
                    reason="Fail",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=self.node_labels[0],
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_STOPPED,
                    node=self.node_labels[1],
                ),
            ]
        )

    def test_node_offline_property_removal(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call(
            communication_list=self._offline_node_communication_list()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + [
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_labels[0],
                    command="api/v1/cluster-property-remove-name/v1",
                    reason="Fail",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.info(
                    report_codes.CIB_CLUSTER_NAME_REMOVED,
                    node=self.node_labels[1],
                ),
            ]
        )

    def test_skip_offline(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(
            communication_list=self._offline_node_communication_list()
        )
        self.fixture_remove_name_prop_call(
            communication_list=self._offline_node_communication_list()
        )
        self.config.env.push_corosync_conf(
            corosync_conf_text=self.fixture_corosync_conf(_FIXTURE_NEW_NAME),
            need_stopped_cluster=True,
            skip_offline_targets=True,
        )

        cluster.rename(
            self.env_assist.get_env(),
            _FIXTURE_NEW_NAME,
            force_flags=[report_codes.SKIP_OFFLINE_NODES],
        )
        self.env_assist.assert_reports(
            [
                fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_labels[0],
                    command="remote/status",
                    reason="Fail",
                ),
                fixture.warn(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=self.node_labels[0],
                ),
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_STOPPED,
                    node=self.node_labels[1],
                ),
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_labels[0],
                    command="api/v1/cluster-property-remove-name/v1",
                    reason="Fail",
                ),
                fixture.info(
                    report_codes.CIB_CLUSTER_NAME_REMOVED,
                    node=self.node_labels[1],
                ),
            ]
        )

    def test_cluster_name_property_remove_invalid_response(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call(output="Invalid json")

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + [
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
            ]
            + [
                fixture.error(report_codes.INVALID_RESPONSE_FORMAT, node=n)
                for n in self.node_labels
            ]
        )

    def test_cluster_name_property_remove_error_report(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call(
            output=json.dumps(
                {
                    "status": "error",
                    "status_msg": None,
                    "report_list": [
                        {
                            "severity": {"level": "ERROR", "force_code": None},
                            "message": {
                                "code": "CIB_XML_MISSING",
                                "message": "CIB XML file cannot be found",
                                "payload": {},
                            },
                            "context": None,
                        }
                    ],
                    "data": "",
                }
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + [
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
            ]
            + [
                fixture.error(
                    report_codes.CIB_XML_MISSING,
                    context=reports.dto.ReportItemContextDto(node=n),
                )
                for n in self.node_labels
            ]
        )

    def test_cluster_name_property_remove_error_status_msg(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call(
            output=json.dumps(
                {
                    "status": "error",
                    "status_msg": "very bad error",
                    "report_list": [],
                    "data": "",
                }
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + [
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
            ]
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=n,
                    command="api/v1/cluster-property-remove-name/v1",
                    reason="very bad error",
                )
                for n in self.node_labels
            ]
        )

    def test_cluster_name_property_remove_error_report_status_msg(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call(
            output=json.dumps(
                {
                    "status": "error",
                    "status_msg": "very bad error",
                    "report_list": [
                        {
                            "severity": {"level": "ERROR", "force_code": None},
                            "message": {
                                "code": "CIB_XML_MISSING",
                                "message": "CIB XML file cannot be found",
                                "payload": {},
                            },
                            "context": None,
                        }
                    ],
                    "data": "",
                }
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + [
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
            ]
            + [
                fixture.error(
                    report_codes.CIB_XML_MISSING,
                    context=reports.dto.ReportItemContextDto(node=n),
                )
                for n in self.node_labels
            ]
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=n,
                    command="api/v1/cluster-property-remove-name/v1",
                    reason="very bad error",
                )
                for n in self.node_labels
            ]
        )

    def test_cluster_name_property_remove_error_no_report_no_message(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call(
            output=json.dumps(
                {
                    "status": "error",
                    "status_msg": None,
                    "report_list": [],
                    "data": "",
                }
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + [
                fixture.info(report_codes.CIB_CLUSTER_NAME_REMOVAL_STARTED),
            ]
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=n,
                    command="api/v1/cluster-property-remove-name/v1",
                    reason="Unknown error",
                )
                for n in self.node_labels
            ]
        )

    def test_corosync_running_on_node(self):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(env=self.runner_env)
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.http.corosync.check_corosync_offline(
            communication_list=[
                {
                    "label": self.node_labels[0],
                    "output": corosync_running_check_response(True),
                }
            ]
            + [{"label": node} for node in self.node_labels[1:]]
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.rename(self.env_assist.get_env(), _FIXTURE_NEW_NAME)
        )
        self.env_assist.assert_reports(
            [
                fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_RUNNING,
                    node=self.node_labels[0],
                ),
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_STOPPED,
                    node=self.node_labels[1],
                ),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_FINISHED_RUNNING,
                    node_list=[self.node_labels[0]],
                ),
            ]
        )


class ClusterRenameCheckGfs2Resources(FixtureMixin, TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.cib_path = os.path.join(settings.cib_dir, "cib.xml")
        self.runner_env = {"CIB_file": self.cib_path}

    def fixture_env(self, resources):
        self.config.fs.exists(self.cib_path, return_value=True)
        self.config.runner.cib.load(
            filename="cib-empty.xml",
            env=self.runner_env,
            resources=resources,
        )
        self.config.corosync_conf.load_content(self.fixture_corosync_conf())
        self.config.env.set_known_nodes(self.node_labels)
        self.config.http.corosync.check_corosync_offline(self.node_labels)
        self.fixture_remove_name_prop_call()
        self.config.env.push_corosync_conf(
            corosync_conf_text=self.fixture_corosync_conf(_FIXTURE_NEW_NAME),
            need_stopped_cluster=True,
        )

    def test_no_gfs2_in_cib(self):
        self.fixture_env(
            f"<resources>{_FIXTURE_GENERIC_FS_RESOURCE}</resources>"
        )
        cluster.rename(
            self.env_assist.get_env(), _FIXTURE_NEW_NAME, force_flags=[]
        )
        self.env_assist.assert_reports(
            self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_gfs2_in_cib(self):
        self.fixture_env(f"<resources>{_FIXTURE_GFS2_FS_RESOURCE}</resources>")
        cluster.rename(
            self.env_assist.get_env(), _FIXTURE_NEW_NAME, force_flags=[]
        )
        self.env_assist.assert_reports(
            [fixture.warn(report_codes.GFS2_LOCK_TABLE_RENAME_NEEDED)]
            + self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_gfs2_cloned_in_cib(self):
        self.fixture_env(
            f"""
                <resources>
                    <clone id="fs-clone">
                        {_FIXTURE_GFS2_FS_RESOURCE}
                    </clone>
                </resources>
            """
        )
        cluster.rename(
            self.env_assist.get_env(), _FIXTURE_NEW_NAME, force_flags=[]
        )
        self.env_assist.assert_reports(
            [fixture.warn(report_codes.GFS2_LOCK_TABLE_RENAME_NEEDED)]
            + self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_dlm_in_cib(self):
        self.fixture_env(f"<resources>{_FIXTURE_DLM_RESOURCE}</resources>")
        cluster.rename(
            self.env_assist.get_env(), _FIXTURE_NEW_NAME, force_flags=[]
        )
        self.env_assist.assert_reports(
            [fixture.warn(report_codes.DLM_CLUSTER_RENAME_NEEDED)]
            + self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_dlm_cloned_in_cib(self):
        self.fixture_env(
            f"""
                <resources>
                    <clone id="dlm-clone">
                        {_FIXTURE_DLM_RESOURCE}
                    </clone>
                </resources>
            """
        )
        cluster.rename(
            self.env_assist.get_env(), _FIXTURE_NEW_NAME, force_flags=[]
        )
        self.env_assist.assert_reports(
            [fixture.warn(report_codes.DLM_CLUSTER_RENAME_NEEDED)]
            + self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )

    def test_dlm_and_gfs2(self):
        self.fixture_env(
            f"""
                <resources>
                    {_FIXTURE_GFS2_FS_RESOURCE}
                    {_FIXTURE_DLM_RESOURCE}
                </resources>
            """
        )
        cluster.rename(
            self.env_assist.get_env(), _FIXTURE_NEW_NAME, force_flags=[]
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(report_codes.DLM_CLUSTER_RENAME_NEEDED),
                fixture.warn(report_codes.GFS2_LOCK_TABLE_RENAME_NEEDED),
            ]
            + self.fixture_corosync_offline_check_reports()
            + self.fixture_remove_name_prop_reports()
        )



import shlex
from dataclasses import replace
from typing import Optional, Sequence

from pcs.cli.common.output import INDENT_STEP, pairs_to_cmd
from pcs.cli.nvset import filter_nvpairs_by_names, nvset_dto_to_lines
from pcs.common.pacemaker.node import CibNodeDto, CibNodeListDto
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.str_tools import indent


def _description_to_lines(desc: Optional[str]) -> list[str]:
    return [f"Description: {desc}"] if desc else []


def _nvsets_to_lines(label: str, nvsets: Sequence[CibNvsetDto]) -> list[str]:
    if nvsets and nvsets[0].nvpairs:
        return nvset_dto_to_lines(nvset=nvsets[0], nvset_label=label)
    return []


def _node_dto_to_nvset_lines(
    node_dto: CibNodeDto, label: str, nvsets: Sequence[CibNvsetDto]
) -> list[str]:
    nvsets_lines = _nvsets_to_lines(label, nvsets)
    if not nvsets_lines:
        return []
    lines = _description_to_lines(node_dto.description)
    lines.extend(nvsets_lines)
    return [f"Node: {node_dto.uname}"] + indent(lines, indent_step=INDENT_STEP)


def config_dto_to_utilization_lines(config_dto: CibNodeListDto) -> list[str]:
    result = []
    for node_dto in config_dto.nodes:
        result.extend(
            _node_dto_to_nvset_lines(
                node_dto, "Utilization", node_dto.utilization
            )
        )
    return result


def config_dto_to_attribute_lines(config_dto: CibNodeListDto) -> list[str]:
    result = []
    for node_dto in config_dto.nodes:
        result.extend(
            _node_dto_to_nvset_lines(
                node_dto, "Attributes", node_dto.instance_attributes
            )
        )
    return result


def _nvsets_to_cmd(
    nvset_cmd: str, node_name: str, nvsets: Sequence[CibNvsetDto]
) -> list[str]:
    if nvsets and nvsets[0].nvpairs:
        node = shlex.quote(node_name)
        options = pairs_to_cmd(
            (nvpair.name, nvpair.value) for nvpair in nvsets[0].nvpairs
        )
        return [f"pcs -- node {nvset_cmd} {node} {options}"]
    return []


def config_dto_to_attribute_cmd(config_dto: CibNodeListDto) -> list[str]:
    commands = []
    for node_dto in config_dto.nodes:
        commands.extend(
            _nvsets_to_cmd(
                "attribute", node_dto.uname, node_dto.instance_attributes
            )
        )
    return commands


def config_dto_to_utilization_cmd(config_dto: CibNodeListDto) -> list[str]:
    commands = []
    for node_dto in config_dto.nodes:
        commands.extend(
            _nvsets_to_cmd("utilization", node_dto.uname, node_dto.utilization)
        )
    return commands


def filter_nodes_by_node_name(
    config_dto: CibNodeListDto, node_name: str
) -> CibNodeListDto:
    return CibNodeListDto(
        nodes=[
            node_dto
            for node_dto in config_dto.nodes
            if node_dto.uname == node_name
        ]
    )


def filter_nodes_nvpairs_by_name(
    config_dto: CibNodeListDto, name: str
) -> CibNodeListDto:
    return CibNodeListDto(
        [
            replace(
                node_dto,
                instance_attributes=filter_nvpairs_by_names(
                    node_dto.instance_attributes, [name]
                ),
                utilization=filter_nvpairs_by_names(
                    node_dto.utilization, [name]
                ),
            )
            for node_dto in config_dto.nodes
        ]
    )



def fixture_property_set(set_id, nvpairs, score=None):
    score_attr = ""
    if score:
        score_attr = f'score="{score}"'
    return (
        f'<cluster_property_set id="{set_id}" {score_attr}>'
        + "".join(
            [
                f'<nvpair id="{set_id}-{name}" name="{name}" value="{value}"/>'
                for name, value in nvpairs.items()
                if value
            ]
        )
        + "</cluster_property_set>"
    )


def fixture_crm_config_properties(set_list, score_list=None):
    return (
        "<crm_config>"
        + "".join(
            [
                fixture_property_set(
                    set_tuple[0],
                    set_tuple[1],
                    score=None if not score_list else score_list[idx],
                )
                for idx, set_tuple in enumerate(set_list)
            ]
        )
        + "</crm_config>"
    )



import json
import shlex
from unittest import TestCase

from pcs.common.interface import dto
from pcs.common.pacemaker.alert import (
    CibAlertDto,
    CibAlertListDto,
    CibAlertRecipientDto,
)
from pcs.common.pacemaker.nvset import CibNvpairDto, CibNvsetDto

from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

# The fixtures are almost the same as in tier0.cli.alert.test_output and they
# should be unified and deduplicated eventually. Currently, it is not possible,
# because the tier0 test tests select, which is not supported by code tested in
# tier1 test.
FIXTURE_ALERTS_DTO = CibAlertListDto(
    [
        CibAlertDto(
            id="alert1",
            path="/path/1",
            description=None,
            recipients=[],
            select=None,
            meta_attributes=[],
            instance_attributes=[],
        ),
        CibAlertDto(
            id="alert2",
            path="/path/2",
            description=None,
            recipients=[
                CibAlertRecipientDto(
                    id="alert2-recipient1",
                    value="test_value_1",
                    description=None,
                    meta_attributes=[],
                    instance_attributes=[],
                ),
                CibAlertRecipientDto(
                    id="alert2-recipient2",
                    value="test_value_2",
                    description="alert2 recipient2 description",
                    meta_attributes=[],
                    instance_attributes=[],
                ),
            ],
            select=None,
            meta_attributes=[],
            instance_attributes=[],
        ),
        CibAlertDto(
            id="alert-all",
            path="/path/all",
            description="alert all options",
            recipients=[
                CibAlertRecipientDto(
                    id="alert-all-recipient",
                    value="value-all",
                    description="all options recipient",
                    meta_attributes=[
                        CibNvsetDto(
                            id="alert-all-recipient-meta_attributes",
                            options={},
                            rule=None,
                            nvpairs=[
                                CibNvpairDto(
                                    id="alert-all-recipient-meta_attributes-aar1m1n",
                                    name="aar1m1n",
                                    value="aar1m1v",
                                ),
                                CibNvpairDto(
                                    id="alert-all-recipient-meta_attributes-aar1m2n",
                                    name="aar1m2n",
                                    value="aar1m2v",
                                ),
                            ],
                        )
                    ],
                    instance_attributes=[
                        CibNvsetDto(
                            id="alert-all-recipient-instance_attributes",
                            options={},
                            rule=None,
                            nvpairs=[
                                CibNvpairDto(
                                    id="alert-all-recipient-instance_attributes-aar1i1n",
                                    name="aar1i1n",
                                    value="aar1i1v",
                                ),
                                CibNvpairDto(
                                    id="alert-all-recipient-instance_attributes-aar1i2n",
                                    name="aar1i2n",
                                    value="aar1i2v",
                                ),
                            ],
                        )
                    ],
                )
            ],
            # eventually, select should not be None, once pcs starts to
            # actually support it in commands for creating alerts
            select=None,
            meta_attributes=[
                CibNvsetDto(
                    id="alert-all-meta_attributes",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(
                            id="alert-all-meta_attributes-aam1n",
                            name="aam1n",
                            value="aam1v",
                        ),
                        CibNvpairDto(
                            id="alert-all-meta_attributes-aam2n",
                            name="aam2n",
                            value="aam2v",
                        ),
                    ],
                )
            ],
            instance_attributes=[
                CibNvsetDto(
                    id="alert-all-instance_attributes",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(
                            id="alert-all-instance_attributes-aai1n",
                            name="aai1n",
                            value="aai1v",
                        ),
                        CibNvpairDto(
                            id="alert-all-instance_attributes-aai2n",
                            name="aai2n",
                            value="aai2v",
                        ),
                    ],
                )
            ],
        ),
    ]
)

FIXTURE_ALERTS_TEXT = """\
Alert: alert1
  Path: /path/1
Alert: alert2
  Path: /path/2
  Recipients:
    Recipient: alert2-recipient1
      Value: test_value_1
    Recipient: alert2-recipient2
      Description: alert2 recipient2 description
      Value: test_value_2
Alert: alert-all
  Description: alert all options
  Path: /path/all
  Recipients:
    Recipient: alert-all-recipient
      Description: all options recipient
      Value: value-all
      Attributes: alert-all-recipient-instance_attributes
        aar1i1n=aar1i1v
        aar1i2n=aar1i2v
      Meta Attributes: alert-all-recipient-meta_attributes
        aar1m1n=aar1m1v
        aar1m2n=aar1m2v
  Attributes: alert-all-instance_attributes
    aai1n=aai1v
    aai2n=aai2v
  Meta Attributes: alert-all-meta_attributes
    aam1n=aam1v
    aam2n=aam2v
"""

FIXTURE_ALERTS_CMD = """\
pcs -- alert create path=/path/1 id=alert1;
pcs -- alert create path=/path/2 id=alert2;
pcs -- alert recipient add alert2 value=test_value_1 id=alert2-recipient1;
pcs -- alert recipient add alert2 value=test_value_2 id=alert2-recipient2 description='alert2 recipient2 description';
pcs -- alert create path=/path/all id=alert-all description='alert all options' options aai1n=aai1v aai2n=aai2v meta aam1n=aam1v aam2n=aam2v;
pcs -- alert recipient add alert-all value=value-all id=alert-all-recipient description='all options recipient' options aar1i1n=aar1i1v aar1i2n=aar1i2v meta aar1m1n=aar1m1v aar1m2n=aar1m2v\
"""


class AlertsConfigJson(TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_empty(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-empty.xml"))
        stdout, stderr, retval = pcs_runner.run(
            ["alert", "config", "--output-format=json"]
        )
        expected = CibAlertListDto([])
        self.assertEqual(
            json.loads(stdout), json.loads(json.dumps(dto.to_dict(expected)))
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-all.xml"))

        stdout, stderr, retval = pcs_runner.run(
            ["alert", "config", "--output-format=json"]
        )
        expected = FIXTURE_ALERTS_DTO
        self.assertEqual(
            json.loads(stdout), json.loads(json.dumps(dto.to_dict(expected)))
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)


class AlertsConfigCmd(TestCase):
    def setUp(self):
        self.old_cib = get_tmp_file("tier1_alert_cmd_old")
        write_file_to_tmpfile(get_test_resource("cib-all.xml"), self.old_cib)
        self.new_cib = get_tmp_file("tier1_alert_cmd_new")
        write_file_to_tmpfile(get_test_resource("cib-empty.xml"), self.new_cib)
        self.old_pcs_runner = PcsRunner(self.old_cib.name)
        self.new_pcs_runner = PcsRunner(self.new_cib.name)
        self.maxDiff = None

    def tearDown(self):
        self.old_cib.close()
        self.new_cib.close()

    def _get_as_json(self, runner):
        stdout, stderr, retval = runner.run(
            ["alert", "config", "--output-format=json"]
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_success(self):
        stdout, stderr, retval = self.old_pcs_runner.run(
            ["alert", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        commands = [
            shlex.split(command)[1:]
            for command in stdout.replace("\\n", "").strip().split(";\n")
        ]

        for cmd in commands:
            stdout, stderr, retval = self.new_pcs_runner.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_as_json(self.new_pcs_runner),
            self._get_as_json(self.old_pcs_runner),
        )


class AlertsConfigText(TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_empty(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-empty.xml"))
        stdout, stderr, retval = pcs_runner.run(
            ["alert", "config", "--output-format=text"]
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-all.xml"))
        stdout, stderr, retval = pcs_runner.run(
            ["alert", "config", "--output-format=text"]
        )
        self.assertEqual(stdout, FIXTURE_ALERTS_TEXT)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success_no_config(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-all.xml"))
        stdout, stderr, retval = pcs_runner.run(["alert"])
        self.assertEqual(stdout, FIXTURE_ALERTS_TEXT)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)



from .cib_to_dto import rule_element_to_dto
from .cib_to_str import RuleToStr
from .expression_part import BoolExpr as RuleRoot
from .in_effect import (
    RuleInEffectEval,
    RuleInEffectEvalDummy,
    RuleInEffectEvalOneByOne,
    get_rule_evaluator,
)
from .parsed_to_cib import export as rule_to_cib
from .parser import RuleParseError, parse_rule
from .tools import is_rsc_expressions_only, is_rsc_expressions_only_dto
from .validator import Validator as RuleValidator



from .types import DeprecatedMessageCode as _D

# DEPRECATED REPORT CODES
# Move the unused report codes to the top of this file and remove
# the ReportItemMessage and CliReportMessageCustom definitions with associated
# report building tests

# Comment structure:
# Removed after <version>[, unused after <version>]

# Use known version number of last release. "Removed" means when was the report
# code deprecated. "Unused" means that the report is no longer produced by PCS
# but the report code and message are still defined.

# Optionally, if the report is being replaced by a different report, a comment
# with the new report code can be added

# Removed after 0.11.6
CANNOT_MOVE_RESOURCE_BUNDLE = _D("CANNOT_MOVE_RESOURCE_BUNDLE")
CANNOT_MOVE_RESOURCE_CLONE = _D("CANNOT_MOVE_RESOURCE_CLONE")

# Removed after 0.11.3
SBD_NOT_INSTALLED = _D("SBD_NOT_INSTALLED")
UNABLE_TO_DETERMINE_USER_UID = _D("UNABLE_TO_DETERMINE_USER_UID")
UNABLE_TO_DETERMINE_GROUP_GID = _D("UNABLE_TO_DETERMINE_GROUP_GID")

# Removed after 0.11.3, unused after 0.11.2
# Replaced by DEFAULTS_CAN_BE_OVERRIDDEN
# Fixed a typo in report code
DEFAULTS_CAN_BE_OVERRIDEN = _D("DEFAULTS_CAN_BE_OVERRIDEN")

# Removed after 0.11.3, unused after 0.10.10
# Replaced by ADD_REMOVE_* reports from new add/remove validator
CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_FOR_NEW_GROUP = _D(
    "CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_FOR_NEW_GROUP"
)
CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP = _D(
    "CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP"
)
CANNOT_GROUP_RESOURCE_ALREADY_IN_THE_GROUP = _D(
    "CANNOT_GROUP_RESOURCE_ALREADY_IN_THE_GROUP"
)
CANNOT_GROUP_RESOURCE_MORE_THAN_ONCE = _D(
    "CANNOT_GROUP_RESOURCE_MORE_THAN_ONCE"
)
CANNOT_GROUP_RESOURCE_NO_RESOURCES = _D("CANNOT_GROUP_RESOURCE_NO_RESOURCES")
CANNOT_GROUP_RESOURCE_NEXT_TO_ITSELF = _D(
    "CANNOT_GROUP_RESOURCE_NEXT_TO_ITSELF"
)

# Removed after 0.11.3, unused after 0.10.8
# Produced only with Pacemaker 1.x
MULTIPLE_SCORE_OPTIONS = _D("MULTIPLE_SCORE_OPTIONS")


# Removed after 0.11.6, unused after 0.11.6
# Replaced by COROSYNC_NOT_RUNNING_CHECK_NODE_STOPPED
# and COROSYNC_NOT_RUNNING_CHECK_NODE_RUNNING
# These reports were replaced as they were to generic (both the codes and the
# messages) and thus didn't convey the required specific information and use
# case
COROSYNC_NOT_RUNNING_ON_NODE = _D("COROSYNC_NOT_RUNNING_ON_NODE")
COROSYNC_RUNNING_ON_NODE = _D("COROSYNC_RUNNING_ON_NODE")

# Removed after 0.11.7, unused after 0.11.7
# Replaced by DUPLICATE_CONSTRAINTS_EXIST and
# pcs.cli.reports.preprocessor.get_duplicate_constraint_exists_preprocessor
DUPLICATE_CONSTRAINTS_LIST = _D("DUPLICATE_CONSTRAINTS_LIST")

# Removed in the first pcs-0.12.x version
# Replaced by COMMAND_MISMATCH_ARGUMENT_TYPE
RESOURCE_STONITH_COMMANDS_MISMATCH = _D("RESOURCE_STONITH_COMMANDS_MISMATCH")

# Removed in the first pcs-0.12.x version
# Pacemaker 3 no longer allows several rules in one constraint. Related code
# was removed from pcs and the report is no longer produced.
CANNOT_ADD_RULE_TO_CONSTRAINT_WRONG_TYPE = _D(
    "CANNOT_ADD_RULE_TO_CONSTRAINT_WRONG_TYPE"
)

# Removed after 0.12.0
# Replaced by STOPPING_RESOURCES, CANNOT_REMOVE_RESOURCES_NOT_STOPPED, and
# STOPPED_RESOURCES_BEFORE_DELETE_CHECK_SKIPPED.
# The 'cib.remove_elements' library command no longer stops resources before
# their deletion.
STOPPING_RESOURCES_BEFORE_DELETING = _D("STOPPING_RESOURCES_BEFORE_DELETING")
STOPPING_RESOURCES_BEFORE_DELETING_SKIPPED = _D(
    "STOPPING_RESOURCES_BEFORE_DELETING_SKIPPED"
)
CANNOT_STOP_RESOURCES_BEFORE_DELETING = _D(
    "CANNOT_STOP_RESOURCES_BEFORE_DELETING"
)



from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
    ensure_unique_args,
    get_rule_str,
    parse_typed_arg,
)
from pcs.cli.reports.output import deprecation_warning
from pcs.cli.reports.preprocessor import (
    get_duplicate_constraint_exists_preprocessor,
)
from pcs.common import (
    const,
    reports,
)
from pcs.common.pacemaker.constraint import get_all_location_constraints_ids
from pcs.common.str_tools import format_list
from pcs.common.types import StringIterable

RESOURCE_TYPE_RESOURCE = "resource"
RESOURCE_TYPE_REGEXP = "regexp"
_RESOURCE_TYPE_MAP = {
    RESOURCE_TYPE_RESOURCE: const.RESOURCE_ID_TYPE_PLAIN,
    RESOURCE_TYPE_REGEXP: const.RESOURCE_ID_TYPE_REGEXP,
}


def remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    # deprecated since pcs-0.11.7
    deprecation_warning(
        "This command is deprecated and will be removed. "
        "Please use 'pcs constraint delete' or 'pcs constraint remove' "
        "instead."
    )
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()
    ensure_unique_args(argv)
    missing_ids = set(argv) - get_all_location_constraints_ids(
        lib.constraint.get_config(evaluate_rules=False)
    )
    if missing_ids:
        raise CmdLineInputError(
            f"Unable to find location constraints: {format_list(missing_ids)}"
        )
    lib.cib.remove_elements(argv)


def _extract_options(
    argv: Argv, options: StringIterable, ignored_options: StringIterable = ()
) -> dict[str, str]:
    result: dict[str, str] = {}
    for argument in argv:
        if "=" not in argument:
            break
        key, value = argument.split("=", 1)
        if key in options:
            result[key] = value
            continue
        if key not in ignored_options:
            break
    return result


def _extract_rule_options(
    argv: Argv, extract_constraint_options: bool = True
) -> tuple[dict[str, str], dict[str, str]]:
    rule_options_def = {"id", "role", "score", "score-attribute"}
    constraint_options_def = {"constraint-id", "resource-discovery"}

    rule_options = _extract_options(
        argv,
        rule_options_def,
        ignored_options=(
            constraint_options_def if extract_constraint_options else set()
        ),
    )
    constraint_options = dict()
    if extract_constraint_options:
        constraint_options = _extract_options(
            argv, constraint_options_def, ignored_options=rule_options_def
        )

    processed_options = set(rule_options_def)
    if extract_constraint_options:
        processed_options |= constraint_options_def
    while (
        argv and "=" in argv[0] and argv[0].split("=")[0] in processed_options
    ):
        argv.pop(0)

    if "constraint-id" in constraint_options:
        constraint_options["id"] = constraint_options["constraint-id"]
        del constraint_options["constraint-id"]

    return rule_options, constraint_options


def create_with_rule(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --force - allow constraint on any resource type, allow duplicate
        constraints
    """
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        raise CmdLineInputError()

    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)

    argv = argv[:]  # eliminate side-effect - do not modify the original argv
    rsc_type, rsc_value = parse_typed_arg(
        argv.pop(0), list(_RESOURCE_TYPE_MAP.keys()), RESOURCE_TYPE_RESOURCE
    )
    if argv[0] == "rule":
        argv.pop(0)
    else:
        raise CmdLineInputError()
    rule_options, constraint_options = _extract_rule_options(argv)

    lib.env.report_processor.set_report_item_preprocessor(
        get_duplicate_constraint_exists_preprocessor(lib)
    )
    lib.constraint_location.create_plain_with_rule(
        _RESOURCE_TYPE_MAP[rsc_type],
        rsc_value,
        get_rule_str(argv) or "",
        rule_options,
        constraint_options,
        force_flags,
    )



import os
from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.lib.commands import cluster_property

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class RemoveClusterName(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.runner_args = [
            settings.cibadmin_exec,
            "--delete-all",
            "--force",
            "--xpath=/cib/configuration/crm_config/cluster_property_set/nvpair[@name='cluster-name']",
        ]
        self.cib_path = os.path.join(settings.cib_dir, "cib.xml")
        self.runner_env = {"CIB_file": self.cib_path}

    def test_success(self):
        self.config.services.is_running("pacemaker", return_value=False)
        self.config.fs.exists(self.cib_path)
        self.config.runner.place(self.runner_args, env=self.runner_env)
        cluster_property.remove_cluster_name(self.env_assist.get_env())

    def test_pacemaker_running(self):
        self.config.services.is_running("pacemaker")
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.remove_cluster_name(
                self.env_assist.get_env()
            )
        )
        self.env_assist.assert_reports(
            [fixture.error(reports.codes.PACEMAKER_RUNNING)]
        )

    def test_no_cib_xml(self):
        self.config.services.is_running("pacemaker", return_value=False)
        self.config.fs.exists(self.cib_path, return_value=False)
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.remove_cluster_name(
                self.env_assist.get_env()
            )
        )
        self.env_assist.assert_reports(
            [fixture.error(reports.codes.CIB_XML_MISSING)]
        )

    def test_runner_failed(self):
        self.config.services.is_running("pacemaker", return_value=False)
        self.config.fs.exists(self.cib_path)
        self.config.runner.place(
            self.runner_args,
            env=self.runner_env,
            returncode=1,
            stdout="foo",
            stderr="bar",
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.remove_cluster_name(
                self.env_assist.get_env()
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CIB_CLUSTER_NAME_REMOVAL_FAILED,
                    reason="bar\nfoo",
                )
            ]
        )



import json
import logging
from typing import (
    Any,
    Optional,
)
from unittest import mock
from urllib.parse import urlencode

from tornado.httpclient import HTTPResponse
from tornado.httputil import HTTPHeaders
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application

from pcs.common import file_type_codes, reports
from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
    TaskResultDto,
)
from pcs.common.async_tasks.types import (
    TaskFinishType,
    TaskKillReason,
    TaskState,
)
from pcs.common.file import RawFileError
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.common.tools import bin_to_str
from pcs.daemon.app import api_v0
from pcs.daemon.async_tasks.scheduler import (
    Scheduler,
    TaskNotFoundError,
)
from pcs.daemon.async_tasks.types import Command
from pcs.lib.auth.types import AuthUser

# Don't write errors to test output.
logging.getLogger("tornado.access").setLevel(logging.CRITICAL)


class MockAuthProvider:
    auth_successful: bool = True
    user = AuthUser("hacluster", ["haclient"])

    def auth_by_token(self, token: str) -> Optional[AuthUser]:
        del token
        return self.user if self.auth_successful else None


class ApiV0Test(AsyncHTTPTestCase):
    """
    Base class for testing API v0, provides useful tools used in tests
    """

    maxDiff = None

    def setUp(self) -> None:
        self.scheduler = mock.AsyncMock(Scheduler)
        self.auth_provider = MockAuthProvider()
        super().setUp()

    def get_app(self) -> Application:
        raise NotImplementedError

    def assert_body(self, val1, val2):
        # TestCase.assertEqual doesn't print full diff when instances of bytes
        # don't match. We want to see the whole diff, so we transform bytes to
        # str. As a bonus, we dont need to specify expected value as bytes.
        def to_str(value):
            return bin_to_str(value) if isinstance(value, bytes) else value

        return self.assertEqual(to_str(val1), to_str(val2))

    def assert_headers(self, headers: HTTPHeaders) -> None:
        banned_headers = {"Server"}
        required_headers = {
            "Cache-Control": "no-store, no-cache",
            "Content-Security-Policy": "frame-ancestors 'self'; default-src 'self'",
            "Pragma": "no-cache",
            "Referrer-Policy": "no-referrer",
            "Strict-Transport-Security": "max-age=63072000",
            "X-Content-Type-Options": "nosniff",
            "X-Frame-Options": "SAMEORIGIN",
            "X-Xss-Protection": "1; mode=block",
        }
        required_real_headers = [
            hdr for hdr in headers.get_all() if hdr[0] in required_headers
        ]
        banned_real_headers = {
            hdr[0] for hdr in headers.get_all() if hdr[0] in banned_headers
        }
        self.assertEqual(
            sorted(required_headers.items()),
            sorted(required_real_headers),
            "Required headers are not present",
        )
        self.assertEqual(
            sorted(banned_real_headers), [], "Banned headers are present"
        )

    def fetch(
        self,
        path: str,
        raise_error: bool = False,
        add_token: bool = True,
        **kwargs: Any,
    ) -> HTTPResponse:
        if add_token:
            # pretends a token was sent in a cookie, so that daemon's
            # authentication mechanism tries to log in a user by the token and
            # calls AuthProvider.auth_by_token
            if "headers" not in kwargs:
                kwargs["headers"] = {}
            if "Cookie" not in kwargs["headers"]:
                kwargs["headers"]["Cookie"] = "token=some-valid-pcsd-token"
            else:
                kwargs["headers"]["Cookie"] += ";token=some-valid-pcsd-token"

        response = super().fetch(
            path,
            raise_error,
            method=("GET" if kwargs.get("body") is None else "POST"),
            **kwargs,
        )
        self.assert_headers(response.headers)
        return response


class BaseApiV0Handler(ApiV0Test):
    """
    Tests of _BaseApiV0Handler class
    """

    url = "/test"
    cmd_name = "test.command"
    cmd_params = {"test": "params"}
    task_ident = "task-ident"
    command_executed = True

    class HandlerForTest(api_v0._BaseApiV0Handler):
        # pylint: disable=protected-access

        def initialize(self, scheduler, auth_provider, cmd_name, cmd_params):
            # pylint: disable=arguments-differ
            # pylint: disable=attribute-defined-outside-init
            super().initialize(scheduler, auth_provider)
            self.cmd_name = cmd_name
            self.cmd_params = cmd_params

        async def _handle_request(self):
            result = await self._process_request(self.cmd_name, self.cmd_params)
            messages = [
                report_item.message.message for report_item in result.reports
            ]
            self.write(
                f"success: {result.success}\n"
                f"result: {result.result}\n"
                f"reports: {messages}\n"
            )

    def get_app(self) -> Application:
        return Application(
            [
                (
                    self.url,
                    self.HandlerForTest,
                    dict(
                        cmd_name=self.cmd_name,
                        cmd_params=self.cmd_params,
                        scheduler=self.scheduler,
                        auth_provider=self.auth_provider,
                    ),
                )
            ]
        )

    def setUp(self):
        super().setUp()
        self.command_dto = CommandDto(
            command_name=self.cmd_name,
            params=self.cmd_params,
            options=CommandOptionsDto(
                effective_username=self.auth_provider.user.username,
                effective_groups=self.auth_provider.user.groups,
            ),
        )
        self.scheduler.new_task.return_value = self.task_ident
        self.addCleanup(self.assert_scheduler_calls)

    def assert_scheduler_calls(self):
        if self.command_executed:
            self.scheduler.new_task.assert_called_once_with(
                Command(self.command_dto, is_legacy_command=True),
                self.auth_provider.user,
            )
            self.scheduler.wait_for_task.assert_called_once_with(
                self.task_ident, self.auth_provider.user
            )

    def test_success(self):
        self.scheduler.wait_for_task.return_value = TaskResultDto(
            task_ident=self.task_ident,
            command=self.command_dto,
            reports=[
                reports.ReportItem.error(
                    reports.messages.StonithUnfencingFailed("an error"),
                ).to_dto()
            ],
            state=TaskState.FINISHED,
            task_finish_type=TaskFinishType.SUCCESS,
            kill_reason=None,
            result="some result of the command",
        )

        response = self.fetch(self.url)
        self.assert_body(
            response.body,
            (
                "success: True\n"
                "result: some result of the command\n"
                "reports: ['Unfencing failed:\\nan error']\n"
            ),
        )
        self.assertEqual(response.code, 200)

    def test_task_not_found(self):
        self.scheduler.wait_for_task.side_effect = TaskNotFoundError(
            self.task_ident
        )

        response = self.fetch(self.url)
        self.assert_body(response.body, "Internal server error")
        self.assertEqual(response.code, 500)

    def test_not_authorized(self):
        self.auth_provider.auth_successful = False
        self.command_executed = False

        response = self.fetch(self.url)
        self.assert_body(response.body, '{"notauthorized":"true"}')
        self.assertEqual(response.code, 401)

    def test_permission_denied(self):
        self.scheduler.wait_for_task.return_value = TaskResultDto(
            task_ident=self.task_ident,
            command=self.command_dto,
            reports=[
                reports.ReportItem.error(
                    reports.messages.NotAuthorized()
                ).to_dto()
            ],
            state=TaskState.FINISHED,
            task_finish_type=TaskFinishType.FAIL,
            kill_reason=None,
            result=None,
        )

        response = self.fetch(self.url)
        self.assert_body(response.body, "Permission denied")
        self.assertEqual(response.code, 403)

    def test_task_timeout(self):
        self.scheduler.wait_for_task.return_value = TaskResultDto(
            task_ident=self.task_ident,
            command=self.command_dto,
            reports=[],
            state=TaskState.FINISHED,
            task_finish_type=TaskFinishType.KILL,
            kill_reason=TaskKillReason.COMPLETION_TIMEOUT,
            result=None,
        )

        response = self.fetch(self.url)
        self.assert_body(response.body, "Task processing timed out")
        self.assertEqual(response.code, 500)

    def test_task_killed(self):
        self.scheduler.wait_for_task.return_value = TaskResultDto(
            task_ident=self.task_ident,
            command=self.command_dto,
            reports=[],
            state=TaskState.FINISHED,
            task_finish_type=TaskFinishType.KILL,
            kill_reason=None,
            result=None,
        )

        response = self.fetch(self.url)
        self.assert_body(response.body, "Task killed")
        self.assertEqual(response.code, 400)

    def test_task_exception(self):
        self.scheduler.wait_for_task.return_value = TaskResultDto(
            task_ident=self.task_ident,
            command=self.command_dto,
            reports=[],
            state=TaskState.FINISHED,
            task_finish_type=TaskFinishType.UNHANDLED_EXCEPTION,
            kill_reason=None,
            result=None,
        )

        response = self.fetch(self.url)
        self.assert_body(response.body, "Unhandled exception")
        self.assertEqual(response.code, 500)


class ApiV0HandlerTest(ApiV0Test):
    """
    Base class for testing _BaseApiV0Handler descendants
    """

    def setUp(self):
        super().setUp()
        self.mock_process_request = mock.AsyncMock()
        process_request_patcher = mock.patch.object(
            # pylint: disable=protected-access
            api_v0._BaseApiV0Handler,
            "_process_request",
            self.mock_process_request,
        )
        process_request_patcher.start()
        self.addCleanup(process_request_patcher.stop)

    def get_app(self) -> Application:
        return Application(
            api_v0.get_routes(self.scheduler, self.auth_provider)
        )

    @staticmethod
    def result_success(result: Any = None) -> api_v0.SimplifiedResult:
        return api_v0.SimplifiedResult(True, result, [])

    @staticmethod
    def result_failure(
        result: Any = None,
        report_items: Optional[list[reports.ReportItemDto]] = None,
    ) -> api_v0.SimplifiedResult:
        return api_v0.SimplifiedResult(False, result, report_items or [])

    def assert_error_with_report(self, url, **kwargs):
        # Test that the handler returns http 400 and report items in body. The
        # actual report items don't matter, we can pick any simple report item.
        self.mock_process_request.return_value = self.result_failure(
            "some error",
            [
                reports.ReportItem.error(
                    reports.messages.StonithUnfencingFailed("an error"),
                    context=reports.ReportItemContext("node1"),
                ).to_dto()
            ],
        )
        response = self.fetch(url, **kwargs)
        self.assert_body(
            response.body, "Error: node1: Unfencing failed:\nan error"
        )
        self.assertEqual(response.code, 400)


class ResourceManageUnmanageMixin:
    body_data = {"resource_list_json": json.dumps(["resource1", "resource2"])}
    command_data = {"resource_or_tag_ids": ["resource1", "resource2"]}

    def test_success(self):
        self.mock_process_request.return_value = self.result_success()
        response = self.fetch(self.url, body=urlencode(self.body_data))
        self.assert_body(response.body, "")
        self.assertEqual(response.code, 200)
        self.mock_process_request.assert_called_once_with(
            self.command_name, self.command_data
        )

    def test_missing_params(self):
        response = self.fetch(self.url)
        self.assert_body(
            response.body, "Required parameters missing: 'resource_list_json'"
        )
        self.assertEqual(response.code, 400)
        self.mock_process_request.assert_not_called()

    def test_json_parse_error(self):
        response = self.fetch(
            self.url, body=urlencode({"resource_list_json": "not a json"})
        )
        self.assert_body(response.body, "Invalid input data format")
        self.assertEqual(response.code, 400)
        self.mock_process_request.assert_not_called()

    def test_failure(self):
        self.assert_error_with_report(self.url, body=urlencode(self.body_data))
        self.mock_process_request.assert_called_once_with(
            self.command_name, self.command_data
        )


class ResourceManageHandler(ResourceManageUnmanageMixin, ApiV0HandlerTest):
    url = "/remote/manage_resource"
    command_name = "resource.manage"


class ResourceUnmanageHandler(ResourceManageUnmanageMixin, ApiV0HandlerTest):
    url = "/remote/unmanage_resource"
    command_name = "resource.unmanage"


class QdeviceNetGetCaCertificateHandler(ApiV0HandlerTest):
    url = "/remote/qdevice_net_get_ca_certificate"

    def test_success(self):
        self.mock_process_request.return_value = self.result_success(
            "certificate data"
        )
        response = self.fetch(self.url)
        self.assert_body(response.body, "certificate data")
        self.assertEqual(response.code, 200)
        self.mock_process_request.assert_called_once_with(
            "qdevice.qdevice_net_get_ca_certificate", {}
        )

    def test_failure(self):
        self.assert_error_with_report(self.url)
        self.mock_process_request.assert_called_once_with(
            "qdevice.qdevice_net_get_ca_certificate", {}
        )


class QdeviceNetSignNodeCertificateHandler(ApiV0HandlerTest):
    url = "/remote/qdevice_net_sign_node_certificate"
    body_data = {
        "certificate_request": "base64 certificate data",
        "cluster_name": "my-cluster",
    }

    def test_success(self):
        self.mock_process_request.return_value = self.result_success(
            "signed certificate data"
        )
        response = self.fetch(self.url, body=urlencode(self.body_data))
        self.assert_body(response.body, "signed certificate data")
        self.assertEqual(response.code, 200)
        self.mock_process_request.assert_called_once_with(
            "qdevice.qdevice_net_sign_certificate_request", self.body_data
        )

    def test_missing_params(self):
        response = self.fetch(self.url)
        self.assert_body(
            response.body,
            "Required parameters missing: 'certificate_request', 'cluster_name'",
        )
        self.assertEqual(response.code, 400)
        self.mock_process_request.assert_not_called()

    def test_failure(self):
        self.assert_error_with_report(self.url, body=urlencode(self.body_data))
        self.mock_process_request.assert_called_once_with(
            "qdevice.qdevice_net_sign_certificate_request", self.body_data
        )


class QdeviceNetClientInitCertificateStorageHandler(ApiV0HandlerTest):
    url = "/remote/qdevice_net_client_init_certificate_storage"
    body_data = {"ca_certificate": "base64 certificate data"}

    def test_success(self):
        self.mock_process_request.return_value = self.result_success()
        response = self.fetch(self.url, body=urlencode(self.body_data))
        self.assert_body(response.body, "")
        self.assertEqual(response.code, 200)
        self.mock_process_request.assert_called_once_with(
            "qdevice.client_net_setup", self.body_data
        )

    def test_missing_params(self):
        response = self.fetch(self.url)
        self.assert_body(
            response.body, "Required parameters missing: 'ca_certificate'"
        )
        self.assertEqual(response.code, 400)
        self.mock_process_request.assert_not_called()

    def test_failure(self):
        self.assert_error_with_report(self.url, body=urlencode(self.body_data))
        self.mock_process_request.assert_called_once_with(
            "qdevice.client_net_setup", self.body_data
        )


class QdeviceNetClientImportCertificateHandler(ApiV0HandlerTest):
    url = "/remote/qdevice_net_client_import_certificate"
    body_data = {"certificate": "base64 certificate data"}

    def test_success(self):
        self.mock_process_request.return_value = self.result_success()
        response = self.fetch(self.url, body=urlencode(self.body_data))
        self.assert_body(response.body, "")
        self.assertEqual(response.code, 200)
        self.mock_process_request.assert_called_once_with(
            "qdevice.client_net_import_certificate", self.body_data
        )

    def test_missing_params(self):
        response = self.fetch(self.url)
        self.assert_body(
            response.body, "Required parameters missing: 'certificate'"
        )
        self.assertEqual(response.code, 400)
        self.mock_process_request.assert_not_called()

    def test_failure(self):
        self.assert_error_with_report(self.url, body=urlencode(self.body_data))
        self.mock_process_request.assert_called_once_with(
            "qdevice.client_net_import_certificate", self.body_data
        )


class QdeviceNetClientDestroyHandler(ApiV0HandlerTest):
    url = "/remote/qdevice_net_client_destroy"

    def test_success(self):
        self.mock_process_request.return_value = self.result_success()
        response = self.fetch(self.url)
        self.assert_body(response.body, "")
        self.assertEqual(response.code, 200)
        self.mock_process_request.assert_called_once_with(
            "qdevice.client_net_destroy", {}
        )

    def test_failure(self):
        self.assert_error_with_report(self.url)
        self.mock_process_request.assert_called_once_with(
            "qdevice.client_net_destroy", {}
        )


class GetConfigsHandler(ApiV0HandlerTest):
    url = "/remote/get_configs"
    request_data = {"cluster_name": "test"}
    command = "pcs_cfgsync.get_configs"

    def test_success(self):
        file_contents = [
            ("foo", "bar"),
            ("foo", None),
            (None, "bar"),
            (None, None),
        ]
        for known_hosts_content, pcs_settings_content in file_contents:
            with self.subTest(
                value=(
                    f"known-hosts: {known_hosts_content};"
                    "pcs_settings.conf: {pcs_settings_content}"
                )
            ):
                self.mock_process_request.reset_mock()
                self.mock_process_request.return_value = self.result_success(
                    SyncConfigsDto(
                        "test",
                        {
                            file_type_codes.PCS_KNOWN_HOSTS: known_hosts_content,
                            file_type_codes.PCS_SETTINGS_CONF: pcs_settings_content,
                        },
                    )
                )
                response = self.fetch(
                    self.url, body=urlencode(self.request_data)
                )
                self.assertEqual(response.code, 200)
                self.assert_body(
                    response.body,
                    json.dumps(
                        {
                            "status": "ok",
                            "cluster_name": "test",
                            "configs": {
                                "known-hosts": {
                                    "type": "file",
                                    "text": known_hosts_content,
                                },
                                "pcs_settings.conf": {
                                    "type": "file",
                                    "text": pcs_settings_content,
                                },
                            },
                        }
                    ),
                )
                self.mock_process_request.assert_called_once_with(
                    self.command, self.request_data
                )

    def test_wrong_cluster_name(self):
        self.mock_process_request.return_value = self.result_failure(
            report_items=[
                reports.ReportItem.error(
                    reports.messages.NodeReportsUnexpectedClusterName("test")
                ).to_dto()
            ],
        )
        response = self.fetch(self.url, body=urlencode(self.request_data))
        self.assertEqual(response.code, 200)
        self.assert_body(
            response.body, json.dumps({"status": "wrong_cluster_name"})
        )
        self.mock_process_request.assert_called_once_with(
            self.command, self.request_data
        )

    def test_cluster_name_parameter_not_provided(self):
        self.mock_process_request.return_value = self.result_failure(
            report_items=[
                reports.ReportItem.error(
                    reports.messages.NodeReportsUnexpectedClusterName("test")
                ).to_dto()
            ],
        )
        response = self.fetch(self.url)
        self.assertEqual(response.code, 200)
        self.assert_body(
            response.body, json.dumps({"status": "wrong_cluster_name"})
        )
        self.mock_process_request.assert_called_once_with(
            self.command, {"cluster_name": ""}
        )

    def test_unable_to_read_corosync_conf(self):
        self.mock_process_request.return_value = self.result_failure(
            report_items=[
                reports.ReportItem.error(
                    reports.messages.FileIoError(
                        file_type_codes.COROSYNC_CONF,
                        RawFileError.ACTION_READ,
                        reason="foo",
                        file_path="bar",
                    )
                ).to_dto()
            ],
        )
        response = self.fetch(self.url, body=urlencode(self.request_data))
        self.assertEqual(response.code, 200)
        self.assert_body(
            response.body, json.dumps({"status": "not_in_cluster"})
        )
        self.mock_process_request.assert_called_once_with(
            self.command, self.request_data
        )

    def test_failure(self):
        self.assert_error_with_report(
            self.url, body=urlencode(self.request_data)
        )
        self.mock_process_request.assert_called_once_with(
            self.command, self.request_data
        )



from .all import (
    CibConstraintsDto,
    get_all_constraints_ids,
    get_all_location_constraints_ids,
    get_all_location_rules_ids,
)
from .colocation import (
    CibConstraintColocationAttributesDto,
    CibConstraintColocationDto,
    CibConstraintColocationSetDto,
)
from .location import (
    CibConstraintLocationAttributesDto,
    CibConstraintLocationDto,
    CibConstraintLocationSetDto,
)
from .order import (
    CibConstraintOrderAttributesDto,
    CibConstraintOrderDto,
    CibConstraintOrderSetDto,
)
from .set import CibResourceSetDto
from .ticket import (
    CibConstraintTicketAttributesDto,
    CibConstraintTicketDto,
    CibConstraintTicketSetDto,
)



from typing import Optional

import dacite

from pcs.common import file_type_codes, reports
from pcs.lib.file.json import JsonParser, JsonParserException
from pcs.lib.interface.config import ParserErrorException, ParserInterface

from .types import KnownHosts


class InvalidFileStructureException(ParserErrorException):
    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg


class Parser(ParserInterface):
    @staticmethod
    def parse(raw_file_data: bytes) -> KnownHosts:
        data = JsonParser.parse(raw_file_data)

        # change the structure of the parsed json, so it can be mapped into our
        # dataclass
        if not isinstance(data, dict):
            raise InvalidFileStructureException("root element is not an object")
        hosts_field = data.get("known_hosts")
        if hosts_field is not None and isinstance(hosts_field, dict):
            for host_name, host_data in hosts_field.items():
                host_data["name"] = host_name

        try:
            return dacite.from_dict(KnownHosts, data)
        except dacite.DaciteError as e:
            raise InvalidFileStructureException(str(e)) from e

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: file_type_codes.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        if isinstance(exception, JsonParserException):
            return JsonParser.exception_to_report_list(
                exception,
                file_type_code,
                file_path,
                force_code,
                is_forced_or_warning,
            )
        if isinstance(exception, InvalidFileStructureException):
            return [
                reports.ReportItem(
                    severity=reports.item.get_severity(
                        force_code,
                        is_forced_or_warning,
                    ),
                    message=reports.messages.ParseErrorInvalidFileStructure(
                        exception.msg, file_type_code, file_path
                    ),
                )
            ]
        raise AssertionError() from exception
